/*
 * OpenConnect (SSL + DTLS) VPN client
 *
 * Copyright © 2008-2012 Intel Corporation.
 * Copyright © 2008 Nick Andrew <nick@nick-andrew.net>
 * Copyright © 2013 John Morrissey <jwm@horde.net>
 *
 * Author: David Woodhouse <dwmw2@infradead.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * version 2.1, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to:
 *
 *   Free Software Foundation, Inc.
 *   51 Franklin Street, Fifth Floor,
 *   Boston, MA 02110-1301 USA
 */

#ifdef HAVE_GETLINE
/* Various BSD systems require this for getline() to be visible */
#define _WITH_GETLINE
#endif

#include <stdio.h>
#ifdef __ANDROID__
#include <android/log.h>
#else
#include <syslog.h>
#endif
#include <stdarg.h>
#include <stdlib.h>
#include <signal.h>
#include <string.h>
#ifdef HAVE_STRINGS_H
#include <strings.h>
#endif
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
//#include <pwd.h>
//#include <sys/utsname.h>
#include <sys/types.h>
//#include <termios.h>
#ifdef LIBPROXY_HDR
#include LIBPROXY_HDR
#endif
//#include <getopt.h>
#include <sys/proc.h>
#include <debug.h>

#include "openconnect-internal.h"

//static void write_progress(void *_vpninfo,
//			   int level, const char *fmt, ...);
static int validate_peer_cert(void *_vpninfo,
			      OPENCONNECT_X509 *peer_cert,
			      const char *reason);
static int process_auth_form(void *_vpninfo,
			     struct oc_auth_form *form);

static int verbose = PRG_INFO;
static int non_inter;
int openconnect_syslog = 0;
extern struct deviosw win_iosw;

//if new feature will be supported, following content can be refered to. Reserve.
#if  0 
static void usage(void)
{
	printf(_("Usage:  openconnect [options] <server>\n"));
	printf(_("Open client for Cisco AnyConnect VPN, version %s\n\n"), openconnect_version_str);
	print_build_opts();
	printf("      --config=CONFIGFILE         %s\n", _("Read options from config file"));
	printf("  -b, --background                %s\n", _("Continue in background after startup"));
	printf("      --pid-file=PIDFILE          %s\n", _("Write the daemon's PID to this file"));
	printf("  -c, --certificate=CERT          %s\n", _("Use SSL client certificate CERT"));
	printf("  -e, --cert-expire-warning=DAYS  %s\n", _("Warn when certificate lifetime < DAYS"));
	printf("  -k, --sslkey=KEY                %s\n", _("Use SSL private key file KEY"));
	printf("  -C, --cookie=COOKIE             %s\n", _("Use WebVPN cookie COOKIE"));
	printf("      --cookie-on-stdin           %s\n", _("Read cookie from standard input"));
	printf("  -d, --deflate                   %s\n", _("Enable compression (default)"));
	printf("  -D, --no-deflate                %s\n", _("Disable compression"));
	printf("      --force-dpd=INTERVAL        %s\n", _("Set minimum Dead Peer Detection interval"));
	printf("  -g, --usergroup=GROUP           %s\n", _("Set login usergroup"));
	printf("  -h, --help                      %s\n", _("Display help text"));
	printf("  -i, --interface=IFNAME          %s\n", _("Use IFNAME for tunnel interface"));
	printf("  -l, --syslog                    %s\n", _("Use syslog for progress messages"));
	printf("  -U, --setuid=USER               %s\n", _("Drop privileges after connecting"));
	printf("      --csd-user=USER             %s\n", _("Drop privileges during CSD execution"));
	printf("      --csd-wrapper=SCRIPT        %s\n", _("Run SCRIPT instead of CSD binary"));
	printf("  -m, --mtu=MTU                   %s\n", _("Request MTU from server"));
	printf("      --base-mtu=MTU              %s\n", _("Indicate path MTU to/from server"));
	printf("  -p, --key-password=PASS         %s\n", _("Set key passphrase or TPM SRK PIN"));
	printf("      --key-password-from-fsid    %s\n", _("Key passphrase is fsid of file system"));
	printf("  -P, --proxy=URL                 %s\n", _("Set proxy server"));
	printf("      --no-proxy                  %s\n", _("Disable proxy"));
	printf("      --libproxy                  %s\n", _("Use libproxy to automatically configure proxy"));
#ifndef LIBPROXY_HDR
	printf("                                  %s\n", _("(NOTE: libproxy disabled in this build)"));
#endif
	printf("  -q, --quiet                     %s\n", _("Less output"));
	printf("  -Q, --queue-len=LEN             %s\n", _("Set packet queue limit to LEN pkts"));
	printf("  -s, --script=SCRIPT             %s\n", _("Shell command line for using a vpnc-compatible config script"));
	printf("                                  %s: \"%s\"\n", _("default"), DEFAULT_VPNCSCRIPT);
	printf("  -S, --script-tun                %s\n", _("Pass traffic to 'script' program, not tun"));
	printf("  -u, --user=NAME                 %s\n", _("Set login username"));
	printf("  -V, --version                   %s\n", _("Report version number"));
	printf("  -v, --verbose                   %s\n", _("More output"));
	printf("      --dump-http-traffic         %s\n", _("Dump HTTP authentication traffic (implies --verbose"));
	printf("  -x, --xmlconfig=CONFIG          %s\n", _("XML config file"));
	printf("      --authgroup=GROUP           %s\n", _("Choose authentication login selection"));
	printf("      --authenticate              %s\n", _("Authenticate only and print login info"));
	printf("      --cookieonly                %s\n", _("Fetch webvpn cookie only; don't connect"));
	printf("      --printcookie               %s\n", _("Print webvpn cookie before connecting"));
	printf("      --cafile=FILE               %s\n", _("Cert file for server verification"));
	printf("      --disable-ipv6              %s\n", _("Do not ask for IPv6 connectivity"));
	printf("      --dtls-ciphers=LIST         %s\n", _("OpenSSL ciphers to support for DTLS"));
	printf("      --no-dtls                   %s\n", _("Disable DTLS"));
	printf("      --no-http-keepalive         %s\n", _("Disable HTTP connection re-use"));
	printf("      --no-passwd                 %s\n", _("Disable password/SecurID authentication"));
	printf("      --no-cert-check             %s\n", _("Do not require server SSL cert to be valid"));
	printf("      --no-xmlpost                %s\n", _("Do not attempt XML POST authentication"));
	printf("      --non-inter                 %s\n", _("Do not expect user input; exit if it is required"));
	printf("      --passwd-on-stdin           %s\n", _("Read password from standard input"));
	printf("      --token-mode=MODE           %s\n", _("Software token type: rsa or totp"));
	printf("      --token-secret=STRING       %s\n", _("Software token secret"));
#ifndef HAVE_LIBSTOKEN
	printf("                                  %s\n", _("(NOTE: libstoken (RSA SecurID) disabled in this build)"));
#endif
#ifndef HAVE_LIBOATH
	printf("                                  %s\n", _("(NOTE: liboath (TOTP) disabled in this build)"));
#endif
	printf("      --reconnect-timeout         %s\n", _("Connection retry timeout in seconds"));
	printf("      --servercert=FINGERPRINT    %s\n", _("Server's certificate SHA1 fingerprint"));
	printf("      --useragent=STRING          %s\n", _("HTTP header User-Agent: field"));
	printf("      --os=STRING                 %s\n", _("OS type (linux,linux-64,mac,win) to report"));
	printf("      --dtls-local-port=PORT      %s\n", _("Set local port for DTLS datagrams"));
	printf("\n");

	helpmessage();
	exit(1);
}
#endif

void write_progress(void *_vpninfo, int level, const char *fmt, ...)
{
	va_list args;
	struct openconnect_info *vpninfo = (struct openconnect_info *)_vpninfo;
	int header[2];

	char	buf[1024];
	
	if (verbose >= level) {		
		va_start(args, fmt);
		vsnprintf(buf, sizeof(buf), (char *)fmt, args);
		va_end(args);
		header[0] = strlen(buf) + 1;
		header[1] = GS_CMD_LOG;
		write(vpninfo->gui_socket_w, header, 8);
		write(vpninfo->gui_socket_w, buf, header[0]);
	}
}

/* Return value:
 *  < 0, on error
 *  = 0, when form was parsed and POST required
 *  = 1, when response was cancelled by user
 */
static int process_auth_form(void *_vpninfo,
			     struct oc_auth_form *form)
{
	struct openconnect_info *vpninfo = _vpninfo;
	struct oc_form_opt *opt;
	char response[1024];
	char *p;

	if (form->banner)
		vpn_progress(vpninfo, PRG_DEBUG, _("%s\n"), form->banner);

	if (form->error) {
		vpn_progress(vpninfo, PRG_ERR, _("%s\n"), form->error);
		goto err;
	}

	if (form->message)
		vpn_progress(vpninfo, PRG_TRACE, _("%s\n"), form->message);

	/* scan for select options first so they are displayed first */
	for (opt = form->opts; opt; opt = opt->next) {
		if (opt->type == OC_FORM_OPT_SELECT) {
			struct oc_form_opt_select *select_opt = (void *)opt;
			struct oc_choice *choice = NULL;
			int i;

			if (!select_opt->nr_choices)
				continue;

			if (vpninfo->authgroup &&
			    !strcmp(opt->name, "group_list")) {
				for (i = 0; i < select_opt->nr_choices; i++) {
					choice = &select_opt->choices[i];

					if (!strcmp(vpninfo->authgroup,
						    choice->label)) {
						opt->value = choice->name;
						break;
					}
				}
				if (!opt->value)
					vpn_progress(vpninfo, PRG_DEBUG,
						     _("Auth choice \"%s\" not available\n"),
						     vpninfo->authgroup);
			}
			if (!opt->value && select_opt->nr_choices == 1) {
				choice = &select_opt->choices[0];
				opt->value = choice->name;
			}
			if (opt->value) {
				select_opt = NULL;
				continue;
			}
			if (non_inter) {
				vpn_progress(vpninfo, PRG_DEBUG,
					     _("User input required in non-interactive mode\n"));
				goto err;
			}
			syslog("%s [", opt->label);
			for (i = 0; i < select_opt->nr_choices; i++) {
				choice = &select_opt->choices[i];
				if (i)
					syslog("|");

				syslog("%s", choice->label);
			}
			syslog("]:");

			if (!fgets(response, sizeof(response), (int)stdin) || !strlen(response))
				goto err;

			p = strchr(response, '\n');
			if (p)
				*p = 0;

			for (i = 0; i < select_opt->nr_choices; i++) {
				choice = &select_opt->choices[i];

				if (!strcmp(response, choice->label)) {
					select_opt->form.value = choice->name;
					break;
				}
			}
			if (!select_opt->form.value) {
				vpn_progress(vpninfo, PRG_DEBUG,
					     _("Auth choice \"%s\" not valid\n"),
					     response);
				goto err;
			}
		}
	}

	for (opt = form->opts; opt; opt = opt->next) {
		if (opt->type == OC_FORM_OPT_TEXT) {
			if (vpninfo->username &&
			    (!strcmp(opt->name, "username")||
			     !strcmp(opt->name, "secondary_username"))) {
				opt->value = strdup(vpninfo->username);
				if (!opt->value)
					goto err;
			} else if (non_inter) {
				vpn_progress(vpninfo, PRG_DEBUG,
					     _("User input required in non-interactive mode\n"));
				goto err;
			} else {
				opt->value = (char *)malloc(80);
				if (!opt->value)
					goto err;

				syslog("%s", opt->label);

				if (!fgets(opt->value, 80, (int)stdin) || !strlen(opt->value))
					goto err;

				p = strchr(opt->value, '\n');
				if (p)
					*p = 0;
			}

		} else if (opt->type == OC_FORM_OPT_PASSWORD) {
			if (vpninfo->password && vpninfo->password[0] > 0
				&& (!strcmp(opt->name, "password") || !strcmp(opt->name, "secondary_password"))) {
				opt->value = strdup(vpninfo->password);
//				vpninfo->password = NULL;
				if (!opt->value)
					goto err;
			} else if (non_inter) {
				vpn_progress(vpninfo, PRG_DEBUG,
					     _("User input required in non-interactive mode\n"));
				goto err;
			} else {
#ifdef WYSE
				vpn_progress(vpninfo, PRG_INFO,
	     				_("Password is needed\n"));
				goto err; 
#else
 				struct termios t;
				opt->value = malloc(80);
				if (!opt->value)
					goto err;

				printf("%s", opt->label);
 
				tcgetattr(0, &t);
				t.c_lflag &= ~ECHO;
				tcsetattr(0, TCSANOW, &t);

				p = fgets(opt->value, 80, stdin);
 
				t.c_lflag |= ECHO;
				tcsetattr(0, TCSANOW, &t);

				printf("\n");

				if (!p || !strlen(opt->value))
					goto err;

				p = strchr(opt->value, '\n');
				if (p)
					*p = 0;
#endif
			}

		}
	}
/*
	if (vpninfo->password) {
		free(vpninfo->password);
		vpninfo->password = NULL;
	}
*/
	return 0;

 err:
	for (opt = form->opts; opt; opt = opt->next) {
		if (opt->value && (opt->type == OC_FORM_OPT_TEXT ||
				   opt->type == OC_FORM_OPT_PASSWORD)) {
			free(opt->value);
			opt->value = NULL;
		}
	}
	return -EINVAL;
}

static void handle_sigusr(int sig)
{
	if (sig == SIGUSR1)
		verbose = PRG_TRACE;
	else if (sig == SIGUSR2)
		verbose = PRG_INFO;
}

void vpn_main(int arg)
{
	struct OPENCONNECTInfo *vi;
	struct OPENCONNECTInfo vidata;
	struct openconnect_info *vpninfo;
	sigaction_t sa;
	int use_syslog = 0;
	char *proxy = NULL;
	int autoproxy = 0;
	int opt;
	char *pidfile = NULL;
	FILE *fp = NULL;
	char *config_arg;
	char *token_str = NULL;
	oc_token_mode_t token_mode = OC_TOKEN_MODE_NONE;
	char *servername, tbuf[12];
	int ret;

	int header[2];


	vpninfo = (struct openconnect_info *)malloc0(sizeof(*vpninfo));
	if (!vpninfo) {
		syslog("OpenConnect: Allocate momery failed\n");
		exit (1);
	}
	memset(vpninfo, 0, sizeof(*vpninfo));

	vpninfo->gui_socket_w = open(GS_NAME"R",O_WRONLY, 0);
	if (vpninfo->gui_socket_w < 0) {
		exit(1);
	}

	vpninfo->gui_socket_r = open(GS_NAME"W",O_RDONLY, 0);
	if (vpninfo->gui_socket_r < 0) {
		exit(1);
	}

	vpninfo->srv_socket_w = open(GS_NAME"RS", O_WRONLY, 0);
	if (vpninfo->gui_socket_r < 0) {
		exit(1);
	}

	vpninfo->srv_socket_r = open(GS_NAME"WS", O_RDONLY, 0);
	if (vpninfo->gui_socket_w < 0) {
		exit(1);
	} else {
		FD_SET(vpninfo->srv_socket_r, &vpninfo->select_rfds);
		if (vpninfo->select_nfds <= vpninfo->srv_socket_r) {
			vpninfo->select_nfds = vpninfo->srv_socket_r + 1;
		}
	}
	
	/* Set up some defaults */
	vpninfo->tun_fd = vpninfo->ssl_fd = vpninfo->dtls_fd = vpninfo->new_dtls_fd  = -1;
	vpninfo->useragent = openconnect_create_useragent("Open AnyConnect VPN Agent");
	vpninfo->time_r = -1000;
	vpninfo->reqmtu = 0;
	vpninfo->deflate = 1;
	vpninfo->dtls_attempt_period = 60;
	vpninfo->max_qlen = 10;
	vpninfo->reconnect_interval = RECONNECT_INTERVAL_MIN;
	vpninfo->reconnect_timeout = 300;
	vpninfo->uid_csd = 0;
	/* We could let them override this on the command line some day, perhaps */
	openconnect_set_reported_os(vpninfo, NULL);
	vpninfo->uid_csd = 0;
	vpninfo->uid_csd_given = 0;
//	vpninfo->validate_peer_cert = validate_peer_cert;
	vpninfo->process_auth_form = process_auth_form;
	vpninfo->cbdata = vpninfo;
	vpninfo->cert_expire_warning = 60 * 86400;
	vpninfo->cancel_fd = -1;
	vpninfo->xmlpost = 1;

	vpninfo->progress = write_progress;
	vpninfo->localname = "localhost";
	vpninfo->ifname = "cscotun0";

	vpn_progress(vpninfo, PRG_INFO, _("OpenConnect %s\n"), openconnect_get_version());

	vpninfo->databuflen = sizeof(vpninfo->data_buf);

	memset(&sa, 0, sizeof(sa));
	sa.sa_handler = handle_sigusr;

	sigaction(SIGUSR1, &sa, NULL);
	sigaction(SIGUSR2, &sa, NULL);


reconnect:

	header[0] = 0;
	header[1] = GS_CMD_CONFIG;
	write(vpninfo->gui_socket_w, header, 8);
	read(vpninfo->gui_socket_r, header, 8);
	if (header[0] != sizeof(vidata) || header[1] != GS_CMD_CONFIG) {
		exit(1);
	}

	vi = &vidata;
	read(vpninfo->gui_socket_r, vi, sizeof(vidata));
	
	ASSERT(vi);
	ASSERT(vi->server[0]);
	
	vpninfo->username = vi->username;
	vpninfo->password= vi->password;
	servername = vi->server;
	vpninfo->dump_http_traffic = 0; 
	/* PRG_INFO: show to user by default; 
	  * PRG_DEBUG: show when details check box is selected
	  * PRG_TRACE: show most detail infomation for developer
	*/
	if (vi->show)
		verbose = PRG_DEBUG; //debug info
	else
		verbose = PRG_INFO;

#ifdef DEBUGGER
	if (vi->showdebug) {
		verbose = PRG_TRACE;
		vpninfo->dump_http_traffic = 1; 
	}
#endif


	if (!vpninfo->hostname) {
		char *url = strdup(servername);

		if (openconnect_parse_url(vpninfo, url)) {
			vpn_progress(vpninfo, PRG_DEBUG, _("Wrong url\n"));
			goto mainloop;
		}
		free(url);
	}

	if (!vpninfo->sslkey)
		vpninfo->sslkey = vpninfo->cert;

	if (proxy && openconnect_set_http_proxy(vpninfo, strdup(proxy)))
		goto mainloop;
	openconnect_reset_st(vpninfo);
	if (!vpninfo->cookie && openconnect_obtain_cookie(vpninfo)) {
		if (vpninfo->csd_scriptname) {
#ifndef WYSE
			unlink(vpninfo->csd_scriptname);
#endif
			vpninfo->csd_scriptname = NULL;
		}
		vpn_progress(vpninfo, PRG_INFO, _("Connect failed\n"));
		openconnect_close_https(vpninfo, 0);
		openconnect_conn_btn_txt(vpninfo, 1);
		openconnect_refresh_status(vpninfo, OPENCONNECT_LOGIN_FAILED);
		vpninfo->status = OPENCONNECT_IDLE;
		goto mainloop;
	}

	if (make_cstp_connection(vpninfo)) {
		vpn_progress(vpninfo, PRG_INFO, _("Creating SSL connection failed\n"));
		openconnect_close_https(vpninfo, 0);
		openconnect_conn_btn_txt(vpninfo, 1);
		goto mainloop;
	}

	if (setup_tun(vpninfo)) {
		vpn_progress(vpninfo, PRG_DEBUG, _("Set up tun device failed\n"));
		openconnect_close_https(vpninfo, 0);
		shutdown_tun(vpninfo);
		openconnect_conn_btn_txt(vpninfo, 1);
		goto mainloop;
	}

#ifdef	WYSE
//	memLogTimestamp(tbuf, sizeof(tbuf));
	vpn_progress(vpninfo, PRG_INFO, _("%sConnection established\n"), tbuf);
#else
	vpn_progress(vpninfo, PRG_INFO,
		     _("Connected %s as %s%s%s, using %s\n"), vpninfo->ifname,
		     vpninfo->vpn_addr?:"",
		     (vpninfo->vpn_addr6 && vpninfo->vpn_addr) ? " + " : "",
		     vpninfo->vpn_addr6 ? : "",
		     (vpninfo->dtls_fd == -1) ?
		     (vpninfo->deflate ? "SSL + deflate" : "SSL")
		     : "DTLS");
#endif

	if (vpninfo->dtls_attempt_period && setup_dtls(vpninfo))
		vpn_progress(vpninfo, PRG_DEBUG, _("Set up DTLS failed; using SSL instead\n"));

	
	openconnect_win_minimize(vpninfo);

mainloop:
	ret = vpn_mainloop(vpninfo);
	if (ret == 1) {
		openconnect_vpninf_free(vpninfo);
		goto reconnect;
	}
	close(vpninfo->gui_socket_w);
	close(vpninfo->gui_socket_r);
	syslog("OpenConnect: exit\n");
	exit (1);
}

proc_t * openconnect(struct OPENCONNECTInfo *vi)
{
	return NULL;
//	return newproc_argc((void (*)())vpn_main, "openconnect", 0, (int)vi);
}

