/*
 * OpenConnect (SSL + DTLS) VPN client
 *
 * Copyright © 2008-2012 Intel Corporation.
 *
 * Author: David Woodhouse <dwmw2@infradead.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * version 2.1, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to:
 *
 *   Free Software Foundation, Inc.
 *   51 Franklin Street, Fifth Floor,
 *   Boston, MA 02110-1301 USA
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <string.h>
#include <signal.h>
#include <fcntl.h>
#include <unistd.h>
#include <netdb.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <arpa/inet.h>
#include <errno.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <net/if_tun.h>
#include <net/route.h>
#include <net/if.h>

#include "openconnect-internal.h"

#define IFF_TUN         LINUX_IFF_TUN
#define IFF_NO_PI     LINUX_IFF_NO_PI
#define TUNSETIFF     LINUX_TUNSETIFF

extern unsigned int Default_gateway;

/*
 * OpenBSD always puts the protocol family prefix onto packets. Other
 * systems let us enable that with the TUNSIFHEAD ioctl, and some of them
 * (e.g. FreeBSD) _need_ it otherwise they'll interpret IPv6 packets as IPv4.
 */
#if defined(__OpenBSD__) || defined(TUNSIFHEAD)
//#define TUN_HAS_AF_PREFIX 1
#endif

static void set_banner(struct openconnect_info *vpninfo)
{
	char *banner, *q;
	const char *p;

	if (!vpninfo->banner || !(banner = (char *)malloc(strlen(vpninfo->banner)+1))) {
		return;
	}
	p = vpninfo->banner;
	q = banner;

	while (*p) {
		if (*p == '%' && isxdigit((int)(unsigned char)p[1]) &&
		    isxdigit((int)(unsigned char)p[2])) {
			*(q++) = unhex(p + 1);
			p += 3;
		} else
			*(q++) = *(p++);
	}
	*q = 0;
	vpn_progress(vpninfo, PRG_INFO, _("\n*** Security Announcement ***\n\n%s\n"), banner);
	free(banner);
}

/*
  * flag: 1 net, 0 host.
  * opt: 1 add, 0 del.
  */
static int route_change(const char *target, const char *netmask, const char *gateway, int flag, int opt)
{
	int ret;
	int fd;
	int cmd;
	struct ortentry ort;
	struct sockaddr_in *addr;

	fd = socket(AF_INET, SOCK_DGRAM, 0);
	if (fd < 0) {
/*		perror(_("open net")); */
		return -EINVAL;
	}
	memset(&ort, 0, sizeof(ort));

	//default is RTF_UP (net) 
	ort.rt_flags = RTF_UP; 
	if (!flag)
		ort.rt_flags |= RTF_HOST;

	if (target) {
		addr = (struct sockaddr_in*) &ort.rt_dst;
		addr->sin_family = AF_INET;
		addr->sin_len = 16;
		addr->sin_addr.s_addr = inet_addr(target);
	}
	if (gateway) {
		addr = (struct sockaddr_in*) &ort.rt_gateway;
		addr->sin_family = AF_INET;
		addr->sin_len = 16;
		addr->sin_addr.s_addr = inet_addr(gateway);
		ort.rt_flags |= RTF_GATEWAY;
	}
	if (netmask && (!(ort.rt_flags & RTF_HOST))) {
		addr = (struct sockaddr_in*) &ort.rt_netmask;
		addr->sin_family = AF_INET;
		addr->sin_len = 16;
		addr->sin_addr.s_addr = inet_addr(netmask);
		ort.rt_subnetmask = inet_addr(netmask);
	}

	if(opt) {
		cmd = SIOCADDRT;
	} else {
		cmd = SIOCDELRT;
	}		
	ret = ioctl(fd, cmd, &ort);
	if(ret < 0)
		syslog(0, PRG_ERR, _("route_change failed, %d\n"), ret);

	close(fd);
	return 0;
}

static int config_tun(struct openconnect_info *vpninfo, int mtu_only)
{
	struct ifreq ifr;
	int net_fd;

	net_fd = socket(AF_INET, SOCK_DGRAM, 0);
	if (net_fd < 0) {
		vpn_progress(vpninfo, PRG_ERR, _("open net error"));
/*		perror(_("open net")); */
		return -EINVAL;
	}
	memset(&ifr, 0, sizeof(ifr));
	strlcpy(ifr.ifr_name, vpninfo->ifname, sizeof(ifr.ifr_name));

	if (!mtu_only) {
		struct sockaddr_in addr;

// On WTOS SIOCSIFFLAGS will be ignored
#ifndef WYSE
		if (ioctl(net_fd, SIOCGIFFLAGS, &ifr) < 0)
			perror(_("SIOCGIFFLAGS"));

		ifr.ifr_flags |= IFF_UP | IFF_POINTOPOINT;
		if (ioctl(net_fd, SIOCSIFFLAGS, &ifr) < 0)
			perror(_("SIOCSIFFLAGS"));
#endif
		memset((void *)&addr, 0, sizeof(struct sockaddr_in));

		addr.sin_family = AF_INET;
		addr.sin_addr.s_addr = inet_addr(vpninfo->vpn_addr);
		addr.sin_len = sizeof(struct sockaddr_in);
		memcpy(&ifr.ifr_addr, &addr, sizeof(addr));
		if (ioctl(net_fd, SIOCSIFADDR, &ifr) < 0) {
/*			perror(_("SIOCSIFADDR")); */
			vpn_progress(vpninfo, PRG_DEBUG, _("set addr failed\n"));
		}
	}

	if (vpninfo->actual_mtu > 0)
		ifr.ifr_mtu = vpninfo->actual_mtu;
	else
		ifr.ifr_mtu = 1412;
	if (ioctl(net_fd, SIOCSIFMTU, &ifr) < 0) {
/*		perror(_("SIOCSIFMTU")); */
		vpn_progress(vpninfo, PRG_DEBUG, _("set mtu failed\n"));		
	}

	close(net_fd);

	return 0;
}

static int process_split_include(struct openconnect_info *vpninfo,
				 const char *route, int operation)
{
	char  network[128];
	char  netmask[128];
	const char *vpnaddr;
	char *slash;
	int ret;

	slash = strchr(route, '/');
	if (!slash) {
		vpn_progress(vpninfo, PRG_DEBUG,
			     _("Discard bad split include: \"%s\"\n"), route);
		return -1;
	}

	*slash = 0;
	strlcpy(network, route, ARRAYSIZE(network));

	/* Put it back how we found it */
	*slash = '/';
	strlcpy(netmask, slash + 1, ARRAYSIZE(netmask));
	
	vpnaddr = vpninfo->vpn_addr;

	if (network[0]) {
		if (operation == 1) {
			ret = route_change(network, netmask, vpnaddr, 1, 1);
			if(ret) {
				vpn_progress(vpninfo, PRG_DEBUG, _("Add vpn_dns route failed\n"));
				return -1;
			}
		} else {
			ret = route_change(network, netmask, vpnaddr, 1, 0);
			if(ret) {
				vpn_progress(vpninfo, PRG_DEBUG, _("Del vpn_dns route failed\n"));		
				return -1;
			}
		}
	}
	
	return 0;
}

static int config_vpn_net_route(struct openconnect_info *vpninfo, int opt)
{

	struct in_addr addr;
	struct in_addr mask;
	char *netaddr;
	int defaultgateway;
	
	if (inet_aton(vpninfo->vpn_addr, &addr) &&
	    inet_aton(vpninfo->vpn_netmask, &mask)) {

		addr.s_addr &= mask.s_addr;
		netaddr = inet_ntoa(addr);

		if (vpninfo->split_includes) {
			route_change(netaddr, vpninfo->vpn_netmask, vpninfo->vpn_addr, 1, opt);
		} else {
			openconnect_get_gateway(vpninfo, &defaultgateway);
			if (opt) // delete original default gateway
				route_change("0.0.0.0", "0.0.0.0", 0, 1, 0);
			route_change("0.0.0.0", "0.0.0.0", vpninfo->vpn_addr, 1, opt);
			if (opt == 0) {	// add back the original default gateway
				addr.s_addr = htonl(defaultgateway);
				route_change("0.0.0.0", "0.0.0.0", inet_ntoa(addr), 1, 1);
			}
		}
	}

	return 0;
}

static int config_vpn_gateway_route(struct openconnect_info *vpninfo, int opt)
{
	char target[128];
	char gateway[128];
	int ret;
	struct in_addr addr;
	int defaultgateway;
	
	if (vpninfo->peer_addr == 0)
		return 0;
	strlcpy(target, inet_ntoa(((struct sockaddr_in *)vpninfo->peer_addr)->sin_addr), ARRAYSIZE(target));
	openconnect_get_gateway(vpninfo, &defaultgateway);
	addr.s_addr = htonl(defaultgateway);
	strlcpy(gateway, inet_ntoa(addr), ARRAYSIZE(gateway));

	ret = route_change(target, "255.255.255.255", gateway, 0, opt);
	if(ret) {
		vpn_progress(vpninfo, PRG_DEBUG, _("set vpn gateway failed\n"));		
		return -1;
	}
	return 0 ;
}

static int do_connect(struct openconnect_info *vpninfo)
{
	int ret;
	char host[80];
	unsigned long vpnaddr, vpnnetaddr, vpnnetmask;
	int fd;
	struct ifreq ifr;
	int i;

	set_banner(vpninfo);

	//set cscotun0	
	ret = config_tun(vpninfo, 0);
	if(ret){
		vpn_progress(vpninfo, PRG_DEBUG, _("set tun failed\n"));		
		return -1;
	}

	// set vpn net route
	if (vpninfo->vpn_netmask) {
		config_vpn_net_route(vpninfo, 1);
	}

	//set vpn gateway route
	config_vpn_gateway_route(vpninfo, 1);

	//set vpn dns route
	for (i = 0; i < 3; i++) {
		if (vpninfo->vpn_dns[i]) {
			ret = route_change(vpninfo->vpn_dns[i], 
					   "255.255.255.255", vpninfo->vpn_addr, 0, 1);

			if(ret) {
				vpn_progress(vpninfo, PRG_DEBUG, 
					_("set vpn_dns route failed\n"));		
				return -1;
			}
		}
	}

	//set split includes route
	if (vpninfo->split_includes) {
		struct split_include *this = vpninfo->split_includes;

		while (this) {
			ret = process_split_include(vpninfo, this->route, 1);
			if (ret)
				return -1;
			this = this->next;
		}
		openconnect_set_dns(vpninfo, 1);
	} else {
		openconnect_set_dns(vpninfo, 2);
	}

	return 0;	
}

static int do_disconnect(struct openconnect_info *vpninfo)
{
	int ret;
	char host[80];
	unsigned long vpnaddr;
	int fd;
	struct ifreq ifr;
	int i;

	//del split includes route
	if (vpninfo->split_includes) {
		struct split_include *this = vpninfo->split_includes;

		while (this) {
			ret = process_split_include(vpninfo, this->route, 0);
			if (ret)
				return -1;
			this = this->next;
		}
	}
	openconnect_set_dns(vpninfo, 0);

	//del vpn dns route
	for (i = 0; i < 3; i++) {
		if (vpninfo->vpn_dns[i]) {
			ret = route_change(vpninfo->vpn_dns[i], 
					"255.255.255.255", vpninfo->vpn_addr, 0, 0);
			if(ret) {
				vpn_progress(vpninfo, PRG_DEBUG, 
					_("del vpn_dns route failed\n"));		
				return -1;
			}
		}
	}	

	//del vpn gateway route
	config_vpn_gateway_route(vpninfo, 0);


	// del vpn net route
	if (vpninfo->vpn_netmask) {
		config_vpn_net_route(vpninfo, 0);
	}

	return 0;
}

/*******************************************
  * config_net(): config tun device and route.
  * reason: 0 disconnect, 1 connect, 2 reconnect.
  */
int config_net(struct openconnect_info *vpninfo, int reason)
{
	int ret;
	
	switch (reason) {
		case 0:
			ret = do_disconnect(vpninfo);
			break;
		case 1:
			ret = do_connect(vpninfo);
			break;
		case 2:
			ret = 0;
			break;
		default:
			vpn_progress(vpninfo, PRG_DEBUG, _("reason not supported\n"));
			ret = -1;
			break;
	}
	return ret;
}

static int os_setup_tun(struct openconnect_info *vpninfo)
{
	int tun_fd = -1;
	struct ifreq ifr;
	int tunerr;

	tun_fd = open("tun", O_RDWR, 0);
	if (tun_fd < 0) {
/*		perror(_("open tun")); */
		vpn_progress(vpninfo, PRG_DEBUG, _("open tun failed\n"));
		return -1;
	}
	/* On WOTS, TUNSETIFF will be ignored */
	memset(&ifr, 0, sizeof(ifr));
	ifr.ifr_flags = IFF_TUN | IFF_NO_PI;
	if (vpninfo->ifname)
		strlcpy(ifr.ifr_name, vpninfo->ifname, sizeof(ifr.ifr_name));
	if (ioctl(tun_fd, TUNSETIFF, (void *) &ifr) < 0) {
		vpn_progress(vpninfo, PRG_DEBUG,
			     _("TUNSETIFF failed\n"));
/*		vpn_progress(vpninfo, PRG_DEBUG,
			     _("TUNSETIFF failed: %s\n"),
			     strerror(errno));*/
		close(tun_fd);
		return -1;
	}
	return tun_fd;
}

/* Set up a tuntap device. */
int setup_tun(struct openconnect_info *vpninfo)
{
	int tun_fd;
	int ret;
	int flag;

	tun_fd = os_setup_tun(vpninfo);
	if (tun_fd < 0)
		return tun_fd;

	ret = config_net(vpninfo, 1);
	if (ret) {
		config_net(vpninfo, 0);		
		close(tun_fd);
		return -1;
	}

	vpninfo->tun_fd = tun_fd;


	if (vpninfo->select_nfds <= tun_fd)
		vpninfo->select_nfds = tun_fd + 1;

	FD_SET(tun_fd, &vpninfo->select_rfds);

	flag = 1;
	ret = ioctl(tun_fd, FIONBIO, &flag);
	if (ret < 0) {
		vpn_progress(vpninfo, PRG_DEBUG, _("set tun FIONBIO failed\n"));
		return ret;
	}

	return 0;
}

int tun_mainloop(struct openconnect_info *vpninfo, int *timeout)
{
	int work_done = 0;
	int prefix_size = 0;
	struct pkt *out_pkt = NULL;
	
#ifdef TUN_HAS_AF_PREFIX
	if (!vpninfo->script_tun)
		prefix_size = sizeof(int);
#endif

	if (FD_ISSET(vpninfo->tun_fd, &vpninfo->select_rfds)) {
		while (1) {
			int len = vpninfo->actual_mtu;

			out_pkt = (struct pkt *)malloc(sizeof(struct pkt) + len);
			if (!out_pkt) {
				vpn_progress(vpninfo, PRG_DEBUG, "Allocation failed\n");
				break;
			}

			len = read(vpninfo->tun_fd, out_pkt->data - prefix_size, len + prefix_size);
			if (len <= prefix_size) {
				free(out_pkt);
				break;
			}
			out_pkt->len = len - prefix_size;

			queue_packet(&vpninfo->outgoing_queue, out_pkt);

			work_done = 1;
			vpninfo->outgoing_qlen++;
			if (vpninfo->outgoing_qlen == vpninfo->max_qlen) {
				vpninfo->drop_packets ++;
				FD_CLR(vpninfo->tun_fd, &vpninfo->select_rfds);
				break;
			}
			vpninfo->out_packets++;
		}
	} else if (vpninfo->outgoing_qlen < vpninfo->max_qlen) {
		FD_SET(vpninfo->tun_fd, &vpninfo->select_rfds);
	}

	/* The kernel returns -ENOMEM when the queue is full, so theoretically
	   we could handle that and retry... but it doesn't let us poll() for
	   the no-longer-full situation, so let's not bother. */
	while (vpninfo->incoming_queue) {
		struct pkt *this = vpninfo->incoming_queue;
		unsigned char *data = this->data;
		int len = this->len;

#ifdef TUN_HAS_AF_PREFIX
		if (!vpninfo->script_tun) {
			struct ip *iph = (void *)data;
			int type;

			if (iph->ip_v == 6)
				type = AF_INET6;
			else if (iph->ip_v == 4)
				type = AF_INET;
			else {
				static int complained = 0;
				if (!complained) {
					complained = 1;
					vpn_progress(vpninfo, PRG_DEBUG,
						     _("Unknown packet (len %d) received: %02x %02x %02x %02x...\n"),
						     len, data[0], data[1], data[2], data[3]);
				}
				free(this);
				continue;
			}
			data -= 4;
			len += 4;
			*(int *)data = htonl(type);
		}
#endif
		vpninfo->incoming_queue = this->next;

		if (write(vpninfo->tun_fd, data, len) < 0) {
			/* Handle death of "script" socket */
			if (vpninfo->script_tun && errno == ENOTCONN) {
				vpninfo->quit_reason = "Client connection terminated";
				return 1;
			}
			vpn_progress(vpninfo, PRG_DEBUG,
				     _("Failed to write incoming packet: %s\n"),
				     strerror(errno));
			vpninfo->drop_packets ++;
		}
		free(this);
		vpninfo->in_packets++;
	}

	/* Work is not done if we just got rid of packets off the queue */
	return work_done;
}

void shutdown_tun(struct openconnect_info *vpninfo)
{

	if(vpninfo->tun_fd > 0) {
		vpn_progress(vpninfo, PRG_DEBUG, _("delete routers\n"));
		config_net(vpninfo, 0);

		vpn_progress(vpninfo, PRG_DEBUG, _("close tun connection\n"));
		close(vpninfo->tun_fd);
		FD_CLR(vpninfo->tun_fd, &vpninfo->select_rfds);
		FD_CLR(vpninfo->tun_fd, &vpninfo->select_wfds);
		FD_CLR(vpninfo->tun_fd, &vpninfo->select_efds);
		vpninfo->tun_fd = -1;
	}

	if (vpninfo->status != OPENCONNECT_DISCONNECTED) {
		openconnect_refresh_status(vpninfo, OPENCONNECT_DISCONNECTED);
	}	

	/* clean up libxml2 resource */
	vpn_progress(vpninfo, PRG_DEBUG,
				     _("Clean up libxml2 resource\n"));	
	xmlCleanupParser();
}

