/*
 * DELL PROPRIETARY INFORMATION
 *
 * This software is confidential.  Dell Inc., or one of its subsidiaries, has
 * supplied this software to you under the terms of a license agreement,
 * nondisclosure agreement or both.  You may not copy, disclose, or use this
 * software except in accordance with those terms.
 *
 * Copyright 2014-2019 Dell Inc. or its subsidiaries.  All Rights Reserved.
 *
 * DELL INC. MAKES NO REPRESENTATIONS OR WARRANTIES
 * ABOUT THE SUITABILITY OF THE SOFTWARE, EITHER EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT. DELL SHALL
 * NOT BE LIABLE FOR ANY DAMAGES SUFFERED BY LICENSEE
 * AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 *
 */

#include <fcntl.h>
#include <time.h>
#include <netinet/in.h>

#include "openconnect-internal.h"

#define	OPENCONNECT_REFRESH	105
#define	OPENCONNECT_REDIAL 	106
#define	REFRESHTIME 5

static const char *openconnect_state_names[] = 
	{	OPENCONNECT_Idle, 
		OPENCONNECT_Connecting, 
		OPENCONNECT_Connected, 
		OPENCONNECT_Discon,
		OPENCONNECT_Idle
	};

void
openconnect_reset_st(struct openconnect_info *vpninfo)
{
	vpninfo->time_s = time(0);
	vpninfo->time_r = time(0);
	vpninfo->in_packets = 0;
	vpninfo->out_packets = 0;
	vpninfo->drop_packets = 0;
	vpninfo->resend_packets = 0;
	vpninfo->status = OPENCONNECT_IDLE;
}

void
openconnect_conn_btn_txt(struct openconnect_info *vpninfo, int recon)
{
	int header[3];

	header[0] = 4;
	header[1] = GS_CMD_BUTTON_TXT;
	header[2] = recon;

	write(vpninfo->gui_socket_w, header, 12);
}

void
openconnect_conn_btn_stat(struct openconnect_info *vpninfo, int enable)
{
	int header[3];

	header[0] = 4;
	header[1] = GS_CMD_BUTTON_ENABLE;
	header[2] = enable;

	write(vpninfo->gui_socket_w, header, 12);
}

void
openconnect_wakeup(struct openconnect_info *vpninfo)
{
	int header[2];

	header[0] = 0;
	header[1] = GS_CMD_WAKEUP;
	write(vpninfo->gui_socket_w, header, 8);
}

char *
uptime(uint ut, char *buf, uint size)
{
	int     days, hrs, mins;

	days = ut / (60*60*24);
	ut -= days * (60*60*24);
	hrs = ut / (60*60);
	ut -= hrs * (60*60);
	mins = ut / 60;
	ut -= mins * 60;

	buf[0] = 0;
	if (days > 0)
#if 0
		slprintf(buf, size, "%d day%s ", days, days > 1 ? "s" :"");
#else
	slprintf(buf, size, "%dd ", days);
#endif
	slprintf(buf, size, "%s%02d:%02d:%02d", buf, hrs, mins, ut);
	return(buf);
}

void
openconnect_refresh(struct openconnect_info *vpninfo)
{
	char	buf[512], uptm[24], addr[64];
	int	header[2];

	uptime(time(0) - vpninfo->time_s, uptm, ARRAYSIZE(uptm));
	if (vpninfo->peer_addr != 0)
		strlcpy(addr, (char *) inet_ntoa(((struct sockaddr_in *)vpninfo->peer_addr)->sin_addr), ARRAYSIZE(addr));
	else
		strlcpy(addr, "N/A", ARRAYSIZE(addr));
	sprintf(buf, OPENCONNECT_STATS, vpninfo->vpn_addr ? vpninfo->vpn_addr : "N/A", addr, 
				uptm, openconnect_state_names[vpninfo->status],
				vpninfo->in_packets, vpninfo->out_packets,
				vpninfo->drop_packets, vpninfo->resend_packets);

	header[0] = strlen(buf) + 1;
	header[1] = GS_CMD_STATUS_STR;
	write(vpninfo->gui_socket_w, header, 8);
	write(vpninfo->gui_socket_w, buf, header[0]);
	//update the last refresh time
	vpninfo->time_r = time(0);
}

void
openconnect_refresh_status(struct openconnect_info *vpninfo, char status)
{
	int	header[3];
	vpninfo->status = status;

	header[0] = 4;
	header[1] = GS_CMD_STATUS;
	header[2] = status;
	write(vpninfo->gui_socket_w, header, 12);

	// selectWindow(vpninfo->win);
	openconnect_refresh(vpninfo);
}


void
openconnect_win_minimize(struct openconnect_info *vpninfo)
{
	int header[2];

	header[0] = 0;
	header[1] = GS_CMD_WIN_MIN;
	write(vpninfo->gui_socket_w, header, 8);
}

void
openconnect_get_gateway(struct openconnect_info *vpninfo, int *gateway)
{
	int header[3];

	header[0] = 0;
	header[1] = GS_CMD_GATEWAY;
	write(vpninfo->gui_socket_w, header, 8);

	read(vpninfo->gui_socket_r, header, 12);
	if (gateway) {
		*gateway = header[2];
	}
}

void
openconnect_set_dns(struct openconnect_info *vpninfo, int append)
{
	int header[6], i;

	header[0] = sizeof(header) - 8;
	header[1] = GS_CMD_DNS;
	header[2] = append;
	
	for (i = 0; i < 3; i++) {
		if (vpninfo->vpn_dns[i])
			header[3 + i] = ntohl(inet_addr(vpninfo->vpn_dns[i]));
		else
			header[3 + i] = 0;
	}
	write(vpninfo->gui_socket_w, header, sizeof(header));
}

int
openconnect_checkcert(struct openconnect_info *vpninfo, char *url)
{
	int header[80];
	int len = strlen(url) + 1;

	len = (len > sizeof(header) - 8) ? sizeof(header) - 8 : len;
	header[0] = len;
	header[1] = GS_CMD_CHECKCERT;
	memset(&header[2], 0, len);
	strncpy((char *) &header[2], url, len - 1);
	write(vpninfo->gui_socket_w, header, 8 + len);

	read(vpninfo->gui_socket_r, header, 12);
	return header[2];
}

int
vpn_gs_process(struct openconnect_info *vpninfo)
{
	char *data;
	int header[3];
	int len;
	int cmd;
	int ret;

	ret = read(vpninfo->srv_socket_r, header, 8);
	if (ret != 8) {
		vpninfo->quit_reason = "close socket";
		FD_CLR(vpninfo->srv_socket_r, &vpninfo->select_rfds);
		return 0;
	}
	len = header[0];
	cmd = header[1];

	data = (char *)malloc(len);
	if (!data)
		return ENOMEM;

	if (read(vpninfo->srv_socket_r, data, len) != len) {
		free(data);
		return EIO;
	}

	switch(cmd) {
	case GS_CMD_CONNECTED:
		header[0] = 4;
		header[1] = GS_CMD_CONNECTED;
		header[2] = (vpninfo->ssl_fd >= 0) ? 1 : 0;
		write(vpninfo->srv_socket_w, header, 12);
		break;
	case GS_CMD_RECONNECT:
		vpninfo->quit_reason = "reconnect req";
		break;
	case GS_CMD_DISCONNECT:
		vpninfo->quit_reason = "disconnect req";
		break;
	case GS_CMD_REFRESH:
		openconnect_refresh(vpninfo);
		break;
	default:
		break;
	}

	free(data);
	return 0;
}

void
openconnect_check_disconn(struct openconnect_info *vpninfo, int ms)
{

	fd_set rfds;
	struct timeval tv;
	time_t  starttime = time(NULL);

	tv.tv_sec = ms / 1000;
	tv.tv_usec = (ms % 1000) * 1000;



	FD_ZERO(&rfds);

	FD_SET(vpninfo->srv_socket_r, &rfds);

	while (select(vpninfo->srv_socket_r + 1, &rfds, 0, 0, &tv) > 0) {
		vpn_gs_process(vpninfo);
		tv.tv_sec = ms / 1000 - (time(NULL) - starttime);
		tv.tv_usec = (ms % 1000) * 1000;
		FD_ZERO(&rfds);
		FD_SET(vpninfo->srv_socket_r, &rfds);
	}
}

void
win_mainloop(struct openconnect_info *vpninfo, fd_set rfds)
{

	if ((time(0) - vpninfo->time_r) >= REFRESHTIME) {
		openconnect_refresh(vpninfo);
		vpninfo->time_r = time(0);
	}

	if (FD_ISSET(vpninfo->srv_socket_r, &rfds)) {
		vpn_gs_process(vpninfo);
	}
	
}

